//
//  db++.C  - simple database engine in C++
//  C++ source code.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "webutil.h"
#include "db++.h"

extern "C"
{
   char *re_comp( const char* );
   int re_exec( const char* );
}

//------------------------------------------------------------------------------

static int Token( char** s, char* t, char sep )
{
   int i;
   char *p = *s;
   if( p == 0 )
   {
      t[0] = 0;
      return 0;
   }
   for( i = 0; ; i++ )
   {
      if( p[i] == sep )
      {
         t[i] = 0; *s = p + i + 1; return 1;
      }
      else if( p[i] == 0 || p[i] == 10 )
      {
         t[i] = 0; *s = 0; return 1;
      }
      else
         t[i] = p[i];
   }
}

//------------------------------------------------------------------------------

int dbInteger::Compare( dbInteger* q )
{
   switch( q->c )
   {
      case C_EQ:   return (i == q->i) ? 1 : 0;
      case C_NE:   return (i != q->i) ? 1 : 0;
      case C_LT:   return (i <  q->i) ? 1 : 0;
      case C_LE:   return (i <= q->i) ? 1 : 0;
      case C_GT:   return (i >  q->i) ? 1 : 0;
      case C_GE:   return (i >= q->i) ? 1 : 0;
      case C_NONE:
      default:     return 1;
   }
}

virtual int dbInteger::Filter( dbField* q )
{
   return Compare( (dbInteger*)q ) == 0 ? -1 : 1;
}

virtual int dbInteger::Set( char* arg, char* val )
{
   if( strcmp( arg, "DBc" ) == 0 )
      c = (DBCompar)atoi( val );
   else return 0;
      return 1;
}

//------------------------------------------------------------------------------

virtual int dbString::Filter( dbField* f )
{
   dbString* q = (dbString*)f;
   if( q->s[0] == 0 ) return 0;
   else
   {
      char tmp[256];
      sprintf( tmp, "%s", q->s );
      lowercase( tmp );
      re_comp( tmp );
      sprintf( tmp, "%s", s );
      lowercase( tmp );
      return re_exec( tmp ) == 0 ? -1 : 1;
   }
}

//------------------------------------------------------------------------------

int dbDate::Compare( dbDate* q )
{
   switch( q->c )
   {
      case C_EQ:   return strcmp( d, q->d ) == 0;
      case C_NE:   return strcmp( d, q->d ) != 0;
      case C_LT:   return strcmp( d, q->d ) <  0;
      case C_LE:   return strcmp( d, q->d ) <= 0;
      case C_GT:   return strcmp( d, q->d ) >  0;
      case C_GE:   return strcmp( d, q->d ) >= 0;
      case C_NONE:
      default:     return 1;
   }
}

virtual int dbDate::Set( char* s )
{
   strncpy( d, s, 6 );
   return 1;
}

virtual int dbDate::Write( char* s )
{
   sprintf( s, "%s", d );
   return 1;
}

virtual int dbDate::Filter( dbField* q )
{
   return Compare( (dbDate*)q ) == 0 ? -1 : 1;
}

virtual int dbDate::Set( char* arg, char* val )
{
   if(      strcmp( arg, "DBy" ) == 0 )
      strncpy( d, val+2, 2 );
   else if( strcmp( arg, "DBm" ) == 0 )
      strncpy( d+2, val, 2 );
   else if( strcmp( arg, "DBd" ) == 0 )
      strncpy( d+4, val, 2 );
   else if( strcmp( arg, "DBc" ) == 0 )
      c = (DBCompar)atoi( val );
   else return 0;
   return 1;
}

//------------------------------------------------------------------------------

virtual int dbOneMany::Set( char* s )
{
   int i;
   char *q = s;
   char r[256];
   for ( i = length; Token( &q, r, ',' ); i++ )
   {
     if( values[i] == 0 )
        values[i] = new char[256];
      sprintf( values[i], "%s", r );
   }
   length = i;
   return 1;
}

virtual int dbOneMany::Write( char* s )
{
   int i;
   s[0] = 0;
   for( i = 0; i < length; i++ )
      sprintf( s+strlen(s), "%s%s", values[i], i+1 < length ? "," : "" );
   return 1;
}

virtual int dbOneMany::Filter( dbField* f )
{
   int i, j, nmatch = 0;
   dbOneMany* q = (dbOneMany*)f;
   if( q->length == 0 ) return 0;
   for( i = 0; i < q->length; i++ )
      for( j = 0; j < length; j++ )
         if( strcmp(  q->values[i], values[j] ) == 0 )
            nmatch++;
   return q->and ? (nmatch == q->length ? 1 : -1) : (nmatch > 0 ? 1 : -1);
}

virtual int dbOneMany::Set( char* arg, char* val )
{
   if( strcmp( arg, "DBa" ) == 0 )
      and = atoi( val );
   else return 0;
   return 1;
}

//------------------------------------------------------------------------------

virtual int dbOneOne::Set( char* s )
{
   int i;
   char *q = s;
   char r[256];
   for ( i = length; Token( &q, r, ',' ); i++ )
   {
     if( values[i] == 0 )
        values[i] = new char[256];
      sprintf( values[i], "%s", r );
   }
   length = i;
   return 1;
}

virtual int dbOneOne::Write( char* s )
{
   int i;
   s[0] = 0;
   for( i = 0; i < length; i++ )
      sprintf( s+strlen(s), "%s%s", values[i], i+1 < length ? "," : "" );
   return 1;
}

virtual int dbOneOne::Filter( dbField* f )
{
   dbOneOne* q = (dbOneOne*)f;
   if( q->length == 0 ) return 0;
   for( int i = 0; i < q->length; i++ )
      if( strcmp(  q->values[i], values[0] ) == 0 )
         return 1;
   return -1;
}

//------------------------------------------------------------------------------

int dbEngine::AddField( DBField* f )
{
   dbField *fd = 0, *fq = 0;
   switch( f->type )
   {
      case STRING:
         fd = new dbString( f->name ); fq = new dbString( f->name );
         break;
      case DATE:
         fd = new dbDate( f->name ); fq = new dbDate( f->name );
         break;
      case NUMBER:
         fd = new dbInteger( f->name ); fq = new dbInteger( f->name );
         break;
      case MASK:
         fd = new dbOneMany( f->name ); fq = new dbOneMany( f->name );
         break;
      case LOOKUP:
         //FILE* fp = fopen( path1 );
         fd = new dbOneOne( f->name );
         fq = new dbOneOne( f->name );
         break;
      default:
         break;
   }
   if( fd )
   {
      data[nfields] =  fd;
      query[nfields] = fq;
      nfields++;
      return 1;
   }
   return 0;
}

int dbEngine::SetField( char* n, char* val, int qry )
{
   int f = FindField( n );
   if( f == -1 )
   {
      char c, m[256];
      if( sscanf( n, "DB%c_%s", &c, m ) == 2 )
      {
         if( (f = FindField(m)) != -1 )
         {
            sprintf( m, "DB%c", c );
            if( qry ) return query[f]->Set( m, val );
            else return data[f]->Set( m, val );
         }
      }
      return 0;
   }
   else
   {
      if( qry ) query[f]->Set( val );
      else data[f]->Set( val );
      return 1;
   }
}

int dbEngine::SetRecord( char* s, int qry )
{
   int i;
   char* p = s;

   for( i = 0; i < nfields; i++ )
   {
      char q[512];
      Token( &p, q, '|' );
      if( qry ) query[i]->Set( q );
      else data[i]->Set( q );
   }
   return i;
}

//------------------------------------------------------------------------------
